import re
from pathlib import Path
from typing import TypedDict, Iterator


class CanvasResource(TypedDict):
    type: str
    name: str
    data: dict | None


class CourseSettings(TypedDict):
    name: str
    code: str
    image: str


class FileData(TypedDict):
    path: str
    canvas_folder: str | None
    lock_at: str | None
    unlock_at: str | None


class ZipFileData(TypedDict):
    zip_file_name: str
    content_folder: str
    additional_files: list[str] | None
    exclude_pattern: str | None
    priority_folder: str | None
    canvas_folder: str | None


class SyllabusData(TypedDict):
    content: str


def iter_keys(text: str) -> Iterator[tuple[str, str, str, str]]:
    for match in re.finditer(fr'@@([^|]+)\|\|([^|]+)\|\|([^@]+)@@', text):
        yield match.group(0), *match.groups()


def get_key(rtype: str, name: str, field: str):
    return f'@@{rtype}||{name}||{field}@@'


class ResourceManager(dict[tuple[str, str], CanvasResource]):

    def add_resource(self, resource: CanvasResource, field: str = None) -> str:
        rtype = resource['type']
        rname = resource['name']
        self[rtype, rname] = resource
        return get_key(rtype, rname, field) if field else None



import json


class MDXCanvasResult:
    def __init__(self, output_file: str):
        self.output_file = output_file
        self.json = {
            "deployed_content": [],
            "content_to_review": [],
            "error": ""
        }

    def add_deployed_content(self, rtype: str, content_name: str, content_url: str = None):
        self.json["deployed_content"].append((rtype, content_name, content_url))

    def add_content_to_review(self, quiz_name: str, link_to_quiz: str):
        self.json["content_to_review"].append([quiz_name, link_to_quiz])

    def get_content_to_review(self):
        return self.json["content_to_review"]

    def add_error(self, error: str):
        self.json["error"] = error

    def output(self):
        if self.output_file:
            with open(self.output_file, 'w') as f:
                f.write(json.dumps(self.json, indent=4))



from canvasapi.course import Course

from ..resources import SyllabusData


class SyllabusObj:
    def __init__(self, course_id: int):
        self.course_id = int
        self.uri = f'/courses/{course_id}/assignments/syllabus'


def deploy_syllabus(course: Course, data: SyllabusData) -> tuple[SyllabusObj, str|None]:
    course.update(course={'syllabus_body': data['content']})
    return SyllabusObj(course.id), None


def lookup_syllabus(course: Course, _: str) -> SyllabusObj:
    return SyllabusObj(course.id)



from bs4 import Tag
from typing import TypedDict, List

from .attributes import Attribute, parse_settings, parse_int
from ..resources import ResourceManager, CanvasResource, get_key


class AssignmentGroupRules(TypedDict, total=False):
    drop_lowest: int
    drop_highest: int
    never_drop: List[int]


def _parse_never_drop_assignments(tag: Tag) -> List[int]:
    never_drop_attr = tag.get('never_drop')
    if not never_drop_attr:
        return []

    never_drop_ids = []
    assignment_names = [name.strip() for name in never_drop_attr.split('|')]

    for assignment_name in assignment_names:
        if assignment_name:
            try:
                assignment_id = get_key('assignment', assignment_name, 'id')
                never_drop_ids.append(assignment_id)
            except Exception as e:
                continue

    return never_drop_ids


def _extract_rules_from_group_data(group_data: dict) -> dict:
    rules: AssignmentGroupRules = {}

    if 'drop_lowest' in group_data and group_data['drop_lowest'] is not None:
        rules['drop_lowest'] = group_data.pop('drop_lowest')

    if 'drop_highest' in group_data and group_data['drop_highest'] is not None:
        rules['drop_highest'] = group_data.pop('drop_highest')

    if 'never_drop' in group_data and group_data['never_drop']:
        rules['never_drop'] = group_data.pop('never_drop')

    if rules:
        group_data['rules'] = rules

    return group_data


class AssignmentGroupTagProcessor:
    """
    Processes <assignment-groups> tags and converts them to Canvas assignment group resources.

    Usage:
        <assignment-groups>
            <group name="Group 1" weight="25" drop_lowest="5" />
            <group name="Group 2" weight="75" drop_highest="3" never_drop="assign1|assign2" />
        </assignment-groups>
    """

    def __init__(self, resource_manager: ResourceManager):
        self._resources = resource_manager

    def __call__(self, group_tag: Tag) -> None:
        self._parse_assignment_group(group_tag)

    def _parse_assignment_group(self, tag: Tag) -> None:
        """
        Parse a single assignment group tag and add it to the resource manager.

        Args:
            tag: The assignment group tag to parse
        """
        never_drop_ids = _parse_never_drop_assignments(tag)

        attribute_fields = [
            Attribute('name', required=True),
            Attribute('weight', new_name='group_weight', parser=parse_int),
            Attribute('drop_lowest', parser=parse_int),
            Attribute('drop_highest', parser=parse_int)
            # TODO: Find additional attributes to support
        ]

        group_data = parse_settings(tag, attribute_fields)
        group_data['never_drop'] = never_drop_ids
        group_data = _extract_rules_from_group_data(group_data)

        assignment_group = CanvasResource(
            type='assignment_group',
            name=group_data['name'],
            data=group_data
        )

        self._resources.add_resource(assignment_group)



import hashlib
import json

from canvasapi.course import Course
from canvasapi.file import File

from .syllabus import SyllabusObj


def update_group_name_to_id(course: Course, data: dict):
    if 'assignment_group' in data:
        data['assignment_group_id'] = _get_group_id(course, data['assignment_group'])


def _get_group_id(course: Course, group_name: str) -> int:
    for group in course.get_assignment_groups():
        if group.name == group_name:
            return group.id

    # Group not found - create it
    # TODO - make assignment_group it's own resource
    #  in which case we would throw an error heres
    return course.create_assignment_group(name=group_name).id


def get_canvas_object(course_getter, attr_name, attr):
    objects = course_getter()
    for obj in objects:
        if obj.__getattribute__(attr_name) == attr:
            return obj
    return None


def get_canvas_uri(canvas_obj):
    if hasattr(canvas_obj, 'html_url'):
        html_url: str = canvas_obj.html_url[len('https://'):]
        domain_end_pos = html_url.find('/')
        return html_url[domain_end_pos:]

    elif isinstance(canvas_obj, File):
        return f'/files/{canvas_obj.id}'

    elif isinstance(canvas_obj, SyllabusObj):
        return canvas_obj.uri

    else:
        raise NotImplementedError(type(canvas_obj))


class ResourceNotFoundException(Exception):
    pass



import re
import textwrap
from xml.etree.ElementTree import Element

import markdown as md
from bs4 import NavigableString, Tag, Comment
from markdown import Extension
from markdown.extensions.codehilite import makeExtension as makeCodehiliteExtension
from markdown.inlinepatterns import BACKTICK_RE, BacktickInlineProcessor

from .inline_math import InlineMathExtension
from ..util import parse_soup_from_xml


class BlackInlineCodeProcessor(BacktickInlineProcessor):
    def handleMatch(self, m: re.Match[str], data: str) -> tuple[Element | str, int, int]:
        el, start, end = super().handleMatch(m, data)
        el.attrib['style'] = 'color: #000000'
        return el, start, end


class BlackInlineCodeExtension(Extension):
    def extendMarkdown(self, md):
        # We use 'backtick' and 190 which are the same values
        # used in markdown.inlinepatterns.py to register the original
        # BacktickInlineCodeProcessor.
        # By reusing the same name, it overrides the original processor with ours
        md.inlinePatterns.register(BlackInlineCodeProcessor(BACKTICK_RE), 'backtick', 190)


def replace_characters(text: str, replacements: dict[str, str]) -> str:
    for old, new in replacements.items():
        text = text.replace(old, new)
    return text


def replace_problematic_characters(text: str, replacements: dict[str, str]) -> str:
    output_lines = []
    lines = iter(text.splitlines())

    # Matches inline code like `code`
    inline_code_pattern = re.compile(r'`([^`]+)`')

    for line in lines:
        if line.strip().startswith('```'):
            output_lines.append(line)
            for code_line in lines:
                output_lines.append(replace_characters(code_line, replacements))
                if code_line.strip().startswith('```'):
                    break
        elif inline_code_pattern.search(line):
            def replacer(match):
                code = match.group(1)
                if any(char in code for char in replacements):
                    return f"`{replace_characters(code, replacements)}`"
                return match.group(0)
            output_lines.append(inline_code_pattern.sub(replacer, line))
        else:
            output_lines.append(line)

    return '\n'.join(output_lines)


def process_markdown_text(text: str) -> str:
    dedented = textwrap.dedent(text)

    html = md.markdown(dedented, extensions=[
        'fenced_code',
        'tables',
        'attr_list',
        'pymdownx.superfences',

        # This embeds the highlight style directly into the HTML
        # instead of using CSS classes
        makeCodehiliteExtension(noclasses=True),

        # This preserves \(...\) inline math expressions
        #  so Canvas will render them with MathJax
        InlineMathExtension(),

        # This forces the color of inline code to be black
        # as a workaround for Canvas's super-ugly default red :P
        BlackInlineCodeExtension(),
        # TODO - can we solve this with baked-in CSS?

        # TODO - add support for tilde => <del> (strikethrough) (look for extension)
        #  or maybe look for a github-flavored-markdown extension
    ])

    return html


def _form_blocks(tag: Tag, excluded: list[str], inline: list[str]) -> tuple[bool, list[Tag]]:
    block_tags = []

    for child in list(tag.children):  # Make a copy because .children will change after the yield
        if isinstance(child, Comment) or child.name in excluded:
            if block_tags:
                yield True, block_tags
            block_tags = []

        elif isinstance(child, NavigableString) or child.name in inline:
            block_tags.append(child)

        else:
            # Some other kind of tag -> breaks up a block
            if block_tags:
                yield True, block_tags
            yield False, [child]
            block_tags = []

    if block_tags:
        yield True, block_tags


def _process_markdown(parent, excluded: list[str], inline: list[str]):
    for needs_markdown, block in _form_blocks(parent, excluded, inline):
        if needs_markdown:
            result = parse_soup_from_xml(
                process_markdown_text(
                    ''.join((str(b) for b in block))
                )
            )
            for tag in block:
                tag.replace_with(result)
        else:
            for tag in block:
                _process_markdown(tag, excluded, inline)


def process_markdown(text: str, excluded: list[str], inline: list[str]) -> str:
    """
    Process Markdown text and return XML text

    This purpose of this function is only the Markdown to XML step
    Custom XML/HTML tags should be handled by the XML processor
    This function simply converts all Markdown formatting to HTML

    This function processes Markdown in ALL XML/HTML tags
    (including nested tags) except those listed in `excluded`.

    :param text: the Markdown text to process
    :param excluded: a list of tag names to exclude; their contents are left untouched
    :returns: The XML/HTML text
    """
    content = replace_problematic_characters(text, {'<': '&lt;'})
    soup = parse_soup_from_xml(content)
    _process_markdown(soup, excluded, inline)
    return str(soup)



from canvasapi.course import Course
from canvasapi.assignment import AssignmentGroup

from .util import get_canvas_object


def _get_group(course: Course, name: str) -> AssignmentGroup:
    return get_canvas_object(course.get_assignment_groups, 'name', name)


def _create_group(course: Course, group_name: str) -> AssignmentGroup:
    return course.create_assignment_group(name=group_name)


def _update_group(group: AssignmentGroup, **kwargs) -> AssignmentGroup:
    return group.edit(**kwargs)


def deploy_group(course: Course, group_data: dict) -> tuple[AssignmentGroup, str | None]:
    if not (group := _get_group(course, group_data['name'])):
        group = _create_group(course, group_data['name'])

    group_data.pop('name', None)
    _update_group(group, **group_data)

    return group, None


lookup_group = _get_group



from bs4 import Tag

from mdxcanvas.resources import ResourceManager, CanvasResource, SyllabusData
from mdxcanvas.util import retrieve_contents


class SyllabusTagProcessor:
    def __init__(self, resources: ResourceManager):
        self._resources = resources

    def __call__(self, tag: Tag):
        syllabus = CanvasResource(
            type='syllabus',
            name='syllabus',
            data=SyllabusData(content=retrieve_contents(tag))
        )
        self._resources.add_resource(syllabus)


from .attributes import Attribute, parse_int, parse_bool, parse_date, parse_settings
from ..util import retrieve_contents
from .quiz_questions import parse_text_question, parse_true_false_question, parse_multiple_choice_question, \
    parse_multiple_answers_question, parse_matching_question, parse_multiple_true_false_question, \
    parse_fill_in_the_blank_question, parse_essay_question, parse_file_upload_question, parse_numerical_question, \
    parse_fill_in_multiple_blanks_question, parse_fill_in_multiple_blanks_filled_answers
from ..resources import ResourceManager, CanvasResource
from bs4 import Tag


class QuizTagProcessor:
    def __init__(self, resources: ResourceManager):
        self._resources = resources
        self.question_types = {
            'text': parse_text_question,
            'true-false': parse_true_false_question,
            'multiple-choice': parse_multiple_choice_question,
            'multiple-answers': parse_multiple_answers_question,
            'matching': parse_matching_question,
            'multiple-tf': parse_multiple_true_false_question,
            'fill-in-the-blank': parse_fill_in_the_blank_question,
            'fill-in-multiple-blanks': parse_fill_in_multiple_blanks_question,
            'fill-in-multiple-blanks-filled-answers': parse_fill_in_multiple_blanks_filled_answers,
            'essay': parse_essay_question,
            'file-upload': parse_file_upload_question,
            'numerical': parse_numerical_question
        }

    def __call__(self, quiz_tag: Tag):
        quiz = {
            "type": "quiz",
            "questions": []
        }
        quiz.update(self._parse_quiz_settings(quiz_tag))

        for tag in quiz_tag.children:
            if not isinstance(tag, Tag):
                continue  # Top-level content is not supported in a quiz tag

            if tag.name == "questions":
                questions = self._parse_questions(tag)
                quiz["questions"].extend(questions)

            elif tag.name == "description":
                quiz["description"] = retrieve_contents(tag)

        info = CanvasResource(
            type='quiz',
            name=quiz['title'],
            data=quiz
        )
        self._resources.add_resource(info)

    def _parse_quiz_settings(self, settings_tag):
        fields = [
            Attribute('title', required=True),
            Attribute('quiz_type', 'assignment'),
            Attribute('assignment_group'),
            Attribute('time_limit', parser=parse_int),
            Attribute('shuffle_answers', False, parse_bool),
            Attribute('hide_results'),  # TODO - should be boolean?
            Attribute('show_correct_answers', True, parse_bool),
            Attribute('show_correct_answers_last_attempt', False, parse_bool),
            Attribute('show_correct_answers_at', parser=parse_date),
            Attribute('hide_correct_answers_at', parser=parse_date),
            Attribute('allowed_attempts', -1, parse_int),
            Attribute('scoring_policy', 'keep_highest'),
            Attribute('one_question_at_a_time', False, parse_bool),
            Attribute('cant_go_back', False, parse_bool),
            Attribute('available_from', parser=parse_date, new_name='unlock_at'),
            Attribute('available_to', parser=parse_date, new_name='lock_at'),
            Attribute('due_at', parser=parse_date),
            Attribute('access_code'),
            Attribute('published', parser=parse_bool),
            Attribute('one_time_results', False, parse_bool),
            Attribute('points_possible')
        ]

        return parse_settings(settings_tag, fields)

    def _parse_questions(self, questions_tag: Tag):
        questions = []
        for question in questions_tag.findAll('question', recursive=False):
            if not (question_type := question.get("type")):
                raise ValueError("Question type not specified.")
            if question_type not in self.question_types:
                raise ValueError(f"Question type {question_type} not supported. Supported types: {', '.join(self.question_types.keys())}")

            parse_tag = self.question_types[question_type]
            result = parse_tag(question)
            questions.extend(result)
        return questions



import csv
import json
from pathlib import Path

import markdowndata
import yaml
from jinja2 import Environment, FileSystemLoader

from ..our_logging import get_logger

logger = get_logger()


def _get_args(args_path: Path, global_args: dict) -> dict | list:
    if args_path.suffix == '.jinja':
        content = _render_template(args_path.read_text(), global_args=global_args)
        args_path = Path(args_path.stem)
    else:
        content = args_path.read_text()

    if args_path.suffix == '.json':
        return json.loads(content)

    elif args_path.suffix == '.csv':
        return list(csv.DictReader(content.splitlines()))

    elif args_path.suffix in ['.md', '.mdd']:
        return markdowndata.loads(content)

    elif args_path.suffix in ['.yaml', '.yml']:
        return yaml.safe_load(content)

    else:
        raise NotImplementedError('Args file of type: ' + args_path.suffix)


def _render_template(
        template: str,
        parent_folder: Path = None,
        args_path: Path = None,
        args: dict | list = None,
        global_args: dict = None,
        templates: list[Path] = None
) -> str:
    loader_paths = [parent_folder, args_path, *(templates or [])]
    loader_paths = [p for p in loader_paths if p is not None]

    env = Environment(
        loader=FileSystemLoader(loader_paths),
    )

    context = {
        "zip": zip,
        "enumerate": enumerate,
        "split_list": lambda x: x.split(";"),
        "read_file": lambda f: (parent_folder / f).absolute().read_text()
    }

    if global_args:
        context |= global_args

    if args:
        context |= {"args": args}

    jj_template = env.from_string(template)
    return jj_template.render(context)


def process_jinja(
        template: str,
        parent_folder: Path = None,
        args_path: Path = None,
        global_args: dict = None,
        templates: list[Path] = None
) -> str:
    if args_path:
        args = _get_args(args_path, global_args)
    else:
        args = {}

    return _render_template(
        template,
        parent_folder=parent_folder,
        args_path=args_path,
        args=args,
        global_args=global_args,
        templates=templates
    )



import logging

logger: logging.Logger = None


def get_logger(course_name: str = None):
    global logger
    if logger is None:
        logger = logging.getLogger('MDXCANVAS')
        logger.setLevel(logging.INFO)

        ch = logging.StreamHandler()
        # ch.setLevel(logging.INFO)

        format_tokens = [
            '%(asctime)s',
            # '%(name)s',
            *([course_name] if course_name is not None else []),
            '%(levelname)s',
            '%(message)s'
        ]
        formatter = logging.Formatter(' - '.join(format_tokens))
        ch.setFormatter(formatter)

        logger.addHandler(ch)

    return logger


def log_warnings(summary: list):
    for warning in summary:
        logger.warning(warning)



from bs4 import Tag

from .attributes import parse_settings, Attribute, parse_bool, parse_date, parse_list, parse_dict, \
    parse_int
from ..util import retrieve_contents
from ..resources import ResourceManager, CanvasResource


class AssignmentTagProcessor:
    def __init__(self, resources: ResourceManager):
        self._resources = resources

    def __call__(self, assignment_tag: Tag):
        fields = [
            Attribute('allowed_attempts', parser=lambda x: -1 if x == 'not_graded' else int(x)),
            Attribute('allowed_extensions', [], parse_list),
            Attribute('annotatable_attachment_id'),  # TODO keep?
            Attribute('assignment_group'),
            Attribute('assignment_overrides'),  # TODO keep?
            Attribute('automatic_peer_reviews', False, parse_bool),
            Attribute('available_from', parser=parse_date, new_name='unlock_at'),
            Attribute('available_to', parser=parse_date, new_name='lock_at'),
            Attribute('due_at', parser=parse_date),
            Attribute('late_due', parser=parse_date),
            Attribute('external_tool_tag_attributes', {}, parser=parse_dict),
            Attribute('final_grader_id'),  # TODO - keep?
            Attribute('grade_group_students_individually', False, parse_bool),
            Attribute('grading_standard_id'),  # TODO - keep?
            Attribute('grading_type', 'points'),
            Attribute('grader_comments_visible_to_graders', False, parse_bool),
            Attribute('grader_count'),
            Attribute('grader_names_visible_to_final_grader', False, parse_bool),
            Attribute('graders_anonymous_to_graders', False, parse_bool),
            Attribute('group_category'),
            Attribute('hide_in_gradebook', False, parse_bool),
            Attribute('integration_data'),  # TODO - keep?
            Attribute('moderated_grading', False, parse_bool),
            Attribute('notify_of_update', False, parse_bool),
            Attribute('omit_from_final_grade', False, parse_bool),
            Attribute('only_visible_to_overrides', False, parse_bool),
            Attribute('peer_reviews', False, parse_bool),
            Attribute('points_possible', parser=parse_int),
            Attribute('position', parser=parse_int),  # TODO - should be int?
            Attribute('published', parser=parse_bool),
            Attribute('quiz_lti'),  # TODO - keep?
            Attribute('submission_types', parser=parse_list),  # TODO - keep?
            Attribute('title', new_name='name', required=True),
            Attribute('turnitin_enabled', False, parse_bool),  # TODO - keep?
            Attribute('turnitin_settings'),  # TODO - keep?
            Attribute('vericite_enabled', False, parse_bool),  # TODO - keep?
        ]

        settings = {
            "type": "assignment",
            "description": retrieve_contents(assignment_tag),
        }

        settings.update(parse_settings(assignment_tag, fields))

        assignment = CanvasResource(
            type='assignment',
            name=settings['name'],
            data=settings
        )
        self._resources.add_resource(assignment)



from bs4 import Tag

from .attributes import parse_settings, Attribute, parse_bool, parse_date
from ..util import retrieve_contents
from ..resources import ResourceManager, CanvasResource


class PageTagProcessor:
    def __init__(self, resources: ResourceManager):
        self._resources = resources

    def __call__(self, page_tag: Tag):
        settings = {
            "type": "page",
            "body": retrieve_contents(page_tag),
        }

        fields = [
            Attribute('title', required=True),
            Attribute('editing_roles', 'teachers'),
            Attribute('notify_of_update', False, parse_bool),
            Attribute('student_todo_at', parser=parse_date),
            Attribute('front_page', False, parse_bool),
            Attribute('published', parser=parse_bool),
            Attribute('publish_at', parser=parse_date)
        ]

        settings.update(parse_settings(page_tag, fields))

        page = CanvasResource(
            type='page',
            name=settings['title'],
            data=settings
        )
        self._resources.add_resource(page)



import textwrap
import warnings

from bs4 import BeautifulSoup, Tag, NavigableString, MarkupResemblesLocatorWarning

# We parse basic strings (no tags) all the time.
# bs4 warns that these might be names or urls.
# We can ignore these warnings.
warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)


def parse_soup_from_xml(text: str) -> BeautifulSoup:
    return BeautifulSoup(text, 'html.parser')


def retrieve_contents(tag: Tag, ignored_child_tag_names: list[str] = ()) -> str:
    """
    Return all the HTML contents of the specified tag
    Excludes the contents of specific sub-tags.
    """
    return textwrap.dedent(
        ''.join(
            str(c)
            for c in tag.contents
            if isinstance(c, NavigableString)
            or (isinstance(c, Tag) and c.name not in ignored_child_tag_names)
        )
    )



import logging
import re
from pathlib import Path
from zipfile import ZipFile, ZipInfo

from .file import deploy_file, lookup_file
from ..our_logging import get_logger
from ..resources import ZipFileData, FileData

logger = get_logger()


def zip_folder(
        folder_path: Path,
        path_to_zip: Path,
        additional_files: list[Path] | None,
        exclude: re.Pattern = None,
        priority_folder: Path = None
):
    """
    Zips a folder, excluding files that match the exclude pattern.
    Items from the standard folder are added to the zip if they are not in the priority folder.
    Items in the priority folder take precedence over items in the standard folder.
    """
    exclude = re.compile(exclude) if exclude else None
    folder_path = folder_path.resolve().absolute()
    logger.debug(f'Zipping {folder_path} to {path_to_zip}')

    priority_files = get_files(priority_folder, exclude, '') if priority_folder else {}
    files = get_files(folder_path, exclude, '')
    if additional_files:
        files.update(get_additional_files(additional_files))

    for zip_name, file in files.items():
        if zip_name not in priority_files:
            priority_files[zip_name] = file
        else:
            logger.debug(f'Preferring {priority_files[zip_name]} over {file}')

    if logger.isEnabledFor(logging.DEBUG):
        file_str = ', '.join(priority_files.keys())
        logger.debug(f'Files for {path_to_zip}: {file_str}')

    write_files(priority_files, path_to_zip)


def get_files(folder_path: Path, exclude: re.Pattern | None, prefix) -> dict[str, Path]:
    if not folder_path.exists():
        raise FileNotFoundError(folder_path)

    files = {}
    for file in folder_path.glob('*'):
        if exclude and exclude.search(file.name):
            logger.debug(f'Excluding {file} from zip')
            continue

        if file.is_dir():
            files.update(get_files(file, exclude, prefix + '/' + file.name))
        else:
            files[prefix + '/' + file.name] = file.absolute()

    return files


def get_additional_files(additional_files: list[Path]) -> dict[str, Path]:
    files = {}
    for file in additional_files:
        if file.is_dir():
            files.update(get_files(file, None, f'/{file.name}'))
        else:
            files[f'/{file.name}'] = file

    return files


def write_files(files: dict[str, Path], path_to_zip: Path):
    with ZipFile(path_to_zip, "w") as zipf:
        for zip_name, file in files.items():
            write_file(file, zip_name.lstrip('/'), zipf)


def make_zip_info(zip_name):
    """
    Ensures that the zip file stays consistent between runs.
    """
    zinfo = ZipInfo(
        zip_name,
        # For consistency, set the time to 1980
        date_time=(1980, 1, 1, 0, 0, 0)
    )
    return zinfo


def write_file(file: Path, zip_name: str, zipf: ZipFile):
    zinfo = make_zip_info(zip_name)
    try:
        with open(file) as f:
            zipf.writestr(zinfo, f.read())
    except UnicodeDecodeError as _:
        logger.debug(f'File {file} encountered a decode error during zip {zipf.filename} creation.')
        with open(file, 'rb') as f:
            zipf.writestr(zinfo, f.read())


def predeploy_zip(zipdata: ZipFileData, tmpdir: Path) -> FileData:
    target_folder = Path(zipdata['content_folder'])

    additional_files = [Path(file) for file in zipdata.get('additional_files') or []]

    pf = zipdata['priority_folder']
    priority_folder = Path(pf) if pf is not None else None
    if priority_folder is not None and not priority_folder.exists():
        raise FileNotFoundError(priority_folder)

    exclude = re.compile(zipdata['exclude_pattern']) if zipdata['exclude_pattern'] is not None else None

    path_to_zip = tmpdir / zipdata['zip_file_name']
    zip_folder(target_folder, path_to_zip, additional_files, exclude, priority_folder)

    file = FileData(
        path=str(path_to_zip),
        canvas_folder=zipdata['canvas_folder']
    )

    return file


deploy_zip = deploy_file
lookup_zip = lookup_file



from canvasapi.course import Course

from ..resources import CourseSettings


class CourseObj:
    def __init__(self, course_id: int):
        self.course_id = int
        self.uri = f'/courses/{course_id}'


def deploy_settings(course: Course, data: CourseSettings) -> tuple[CourseObj, str | None]:

    course.update(course={
        'name': data['name'],
        'course_code': data['code'],
        'image_id': int(data['image']) if data.get('image') else None
    })
    return CourseObj(course.id), None


def lookup_settings(course: Course, _: str) -> CourseObj:
    return CourseObj(course.id)



import hashlib
import json
import requests
from pathlib import Path
from tempfile import TemporaryDirectory

from canvasapi.course import Course

from .file import get_file, deploy_file
from ..resources import FileData

MD5_FILE_NAME = '_md5sums.json'


def compute_md5(obj: dict):
    if 'path' in obj:  # e.g. FileData
        path = Path(obj['path'])
        hashable = path.name.encode() + path.read_bytes()
    else:
        hashable = json.dumps(obj).encode()

    return hashlib.md5(hashable).hexdigest()


class MD5Sums:
    def __init__(self, course: Course):
        self._course = course

    def _download_md5s(self):
        md5_file = get_file(self._course, MD5_FILE_NAME)
        if md5_file is None:
            self._md5s = {}
        else:
            self._md5s = {
                tuple(k.split('|')): v
                for k, v in json.loads(requests.get(md5_file.url).text).items()
            }

    def _save_md5s(self):
        with TemporaryDirectory() as tmpdir:
            tmpfile = Path(tmpdir) / MD5_FILE_NAME
            tmpfile.write_text(json.dumps({'|'.join(k): v for k, v in self._md5s.items()}))
            deploy_file(self._course, FileData(
                path=str(tmpfile.absolute()),
                canvas_folder="_md5s"
            ))

    def get(self, item, *args, **kwargs):
        return self._md5s.get(item, *args, **kwargs)

    def __getitem__(self, item):
        # Act like a dictionary
        return self._md5s[item]

    def __setitem__(self, key, value):
        self._md5s[key] = value

    def __enter__(self):
        self._download_md5s()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._save_md5s()



import argparse
import json
import os
from pathlib import Path

from canvasapi import exceptions
from canvasapi.paginated_list import PaginatedList

from ..main import CourseInfo, get_course, load_config

from ..our_logging import get_logger


def get_item_name(item):
    if hasattr(item, 'title'):
        return item.title
    elif hasattr(item, 'name'):
        return item.name
    elif hasattr(item, 'display_name'):
        return item.display_name
    elif hasattr(item, 'filename'):
        return item.filename
    else:
        return str(item)

def get_item_type(item):
    if hasattr(item, 'is_quiz_assignment'):
        if item.is_quiz_assignment:
            return 'Quiz'
        else:
            return 'Assignment'


def delete_item(item, item_name):
    logger = get_logger()

    try:
        item.delete()
    except exceptions.BadRequest as e:
        if "Can't delete the root folder" in str(e):
            logger.info(f'Skipping root folder: {item_name}')
        else:
            logger.warning(f'Failed to delete {item.type}: {item_name}')


def remove(items: PaginatedList, item_type=None):
    logger = get_logger()

    for item in items:
        # Conditions to help with the removal of files and folders
        if hasattr(item, 'get_folders'):
            sub_folders = item.get_folders()
            if item.parent_folder_id is None:
                continue
            remove(sub_folders, 'Folder')
        if hasattr(item, 'get_files'):
            files = item.get_files()
            remove(files, 'File')

        if item_type is None:
            item_type = get_item_type(item)
            item_name = get_item_name(item)
            logger.info(f'Deleting {item_type}: {item_name}')
            item_type = None
        else:
            item_name = get_item_name(item)
            logger.info(f'Deleting {item_type}: {item_name}')
        delete_item(item, item_name)


def main(
        canvas_api_token: str,
        course_info: CourseInfo
):

    course = get_course(canvas_api_token,
                        course_info['CANVAS_API_URL'],
                        course_info['CANVAS_COURSE_ID'])

    logger = get_logger(course.name)
    logger.info('Connecting to Canvas...')

    course.update(course={'syllabus_body': ''})
    logger.info('Deleting Syllabus')

    assignments = course.get_assignments()
    remove(assignments) if len(list(assignments)) > 0 else None

    assignment_groups = course.get_assignment_groups()
    remove(assignment_groups, 'Assignment Group') if len(list(assignment_groups)) > 0 else None

    pages = course.get_pages()
    remove(pages, 'Page') if len(list(pages)) > 0 else None

    modules = course.get_modules()
    remove(modules, 'Module') if len(list(modules)) > 0 else None

    files = course.get_folders()
    remove(files, 'Folder') if len(list(files)) > 0 else None


def entry():
    parser = argparse.ArgumentParser()
    parser.add_argument("--course-info", type=Path)
    parser.add_argument('-y', action='store_true')
    args = parser.parse_args()

    course_settings = load_config(args.course_info)

    api_token = os.environ.get("CANVAS_API_TOKEN")
    if api_token is None:
        raise ValueError("Please set the CANVAS_API_TOKEN environment variable")

    logger = get_logger()

    if not args.y:
        confirm = input('Are you sure you want to delete all course content? (y/[n]): ')
        if confirm.lower() != 'y':
            logger.info('Exiting...')
            return

    main(
        canvas_api_token=api_token,
        course_info=course_settings
    )


if __name__ == '__main__':
    entry()
    


import textwrap
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Callable, Union
from typing import Protocol, TypeAlias

import pytz
from bs4 import BeautifulSoup
from bs4.element import Tag, NavigableString
from jinja2 import Environment


ResourceExtractor: TypeAlias = Callable[[str], tuple[str, list]]


def order_elements(element: dict) -> OrderedDict:
    new_list = []
    for key, value in element.items():
        if isinstance(value, dict):
            new_list.append((key, order_elements(value)))
        else:
            new_list.append((key, value))
    return OrderedDict(sorted(element.items(), key=lambda x: x[0]))


def get_corrects(question_tag):
    corrects = question_tag.select('correct')
    return corrects


def get_correct_comments(question_tag):
    feedback = question_tag.select('correct-comments')
    return get_text_contents(feedback[0]) if feedback else None


def get_incorrect_comments(question_tag):
    feedback = question_tag.select('incorrect-comments')
    return get_text_contents(feedback[0]) if feedback else None


def get_points(question_tag, default=1):
    points = question_tag.get("points", default)
    try:
        return int(points)
    except ValueError:
        print("Invalid points value: " + points)
        return default


def get_answers(question_tag):
    return question_tag.select('correct, incorrect')


def string_is_date(date: str):
    # For templating. The string might not be a date yet.
    # Once the template arguments are filled in, we will apply make_iso.
    if date.startswith("{") or "due" in date.lower() or "lock" in date.lower():
        return False
    has_digit = False
    for d in range(10):
        if f"{d}" in date:
            has_digit = True
    return has_digit


def make_iso(date: datetime | str | None, time_zone: str) -> str:
    if isinstance(date, datetime):
        return datetime.isoformat(date)
    elif isinstance(date, str):
        # Check if the string is already in ISO format
        try:
            return datetime.isoformat(datetime.fromisoformat(date))
        except ValueError:
            pass

        try_formats = [
            "%b %d, %Y, %I:%M %p",
            "%b %d %Y %I:%M %p",
            "%Y-%m-%dT%H:%M:%S%z"
        ]
        for format_str in try_formats:
            try:
                parsed_date = datetime.strptime(date, format_str)
                break
            except ValueError:
                pass
        else:
            raise ValueError(f"Invalid date format: {date}")

        # Convert the parsed datetime object to the desired timezone
        to_zone = pytz.timezone(time_zone)
        parsed_date = parsed_date.replace(tzinfo=None)  # Remove existing timezone info
        parsed_date = parsed_date.astimezone(to_zone)
        return datetime.isoformat(parsed_date)
    else:
        raise TypeError("Date must be a datetime object or a string")


def get_text_contents(tag, children_tag_names: list[str] = ()):
    """
    Typically, the body of a tag is found at contents[0], and sub-tags (like answers) are found later.
    However, images sometimes separate the text into multiple parts.
    This function joins the text and images together.
    """
    return textwrap.dedent("".join(
        [str(c) for c in tag.contents if
         isinstance(c, NavigableString) or (isinstance(c, Tag) and c.name not in children_tag_names)]))


question_types = [
    'calculated_question',
    'essay_question',
    'file_upload_question',
    'fill_in_multiple_blanks_question',
    'matching_question',
    'multiple_answers_question',
    'multiple_choice_question',
    'multiple_dropdowns_question',
    'numerical_question',
    'short_answer_question',
    'text_only_question',
    'true_false_question'
]


class TFConverter:
    @staticmethod
    def process(correct_incorrect_tag, markdown_processor: ResourceExtractor):
        is_true = correct_incorrect_tag.name == "correct"
        question_text, resources = markdown_processor(
            get_text_contents(correct_incorrect_tag, ["correct-comments", "incorrect-comments"]))
        question = {
            "question_text": question_text,
            "question_type": 'true_false_question',
            "points_possible": get_points(correct_incorrect_tag),
            "correct_comments": get_correct_comments(correct_incorrect_tag),
            "incorrect_comments": get_incorrect_comments(correct_incorrect_tag),
            "answers": [
                {
                    "answer_text": "True",
                    "answer_weight": 100 if is_true else 0
                },
                {
                    "answer_text": "False",
                    "answer_weight": 0 if is_true else 100
                }
            ]
        }
        return question, resources


class Processor(Protocol):
    """
    A processor takes a question tag and a Markdown processor
      returning question(s) and a list of resources
    """

    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor) -> Union[
        tuple[list[dict], list], tuple[dict, list]]: ...


class AttributeAdder:
    def __init__(self, settings: dict, settings_tag: Tag, parser):
        self.settings = settings
        self.settings_tag = settings_tag
        self.parser = parser

    def __call__(self, attribute, default=None, new_name=None, formatter=None, typ=None):
        if attribute in self.settings_tag.attrs or default is not None:
            value = self.settings_tag.get(attribute, default)
            if formatter:
                value = formatter(value)
            if typ:
                value = self.parser.parse(value, typ)
            self.settings[new_name if new_name else attribute] = value


def process(processor: Processor, question_tag, markdown_processor: ResourceExtractor):
    question, resources = processor.process(question_tag, markdown_processor)
    return question, resources


class TrueFalseProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        answers = get_answers(question_tag)

        check_answer_size(answers, 1, "True/False question must have exactly 1 answer!")

        question, resources = process(TFConverter(), answers[0], markdown_processor)
        if not get_points(answers[0], 0):
            points = get_points(question_tag)
            question["points_possible"] = points
        if not question["correct_comments"]:
            question["correct_comments"] = get_correct_comments(question_tag)
        if not question["incorrect_comments"]:
            question["incorrect_comments"] = get_incorrect_comments(question_tag)
        return question, resources


class MultipleTrueFalseProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        header_question_text, resources = markdown_processor(
            get_text_contents(question_tag, ["correct", "incorrect", "correct-comments", "incorrect-comments"]))
        header_question = {
            "question_text": header_question_text,
            "question_type": 'text_only_question',
        }
        questions = [header_question]
        for answer in get_answers(question_tag):
            tf_question, res = process(TFConverter(), answer, markdown_processor)
            questions.append(tf_question)
            resources.extend(res)
        return questions, resources


def check_answer_size(answers: list, num, explanation):
    if num is not None and len(answers) != num:
        raise Exception(f"{explanation}\n"
                        "Answers: " + str(answers))


class MultipleCommonProcessor:
    question_type: str
    num_correct: Union[int, None]

    def process(self, question_tag, markdown_processor: ResourceExtractor):
        corrects = get_corrects(question_tag)
        check_answer_size(corrects, self.num_correct,
                          f"{self.question_type} questions must have exactly {self.num_correct} correct answer!")

        question_text, resources = markdown_processor(
            get_text_contents(question_tag, ["correct", "incorrect", "correct-comments", "incorrect-comments"]))
        answers = []
        for answer in get_answers(question_tag):
            answer_html, res = markdown_processor(get_text_contents(answer))
            answers.append((True if answer in corrects else False, answer_html))
            resources.extend(res)

        question = {
            "question_text": question_text,
            "question_type": self.question_type,
            "points_possible": get_points(question_tag),
            "correct_comments": get_correct_comments(question_tag),
            "incorrect_comments": get_incorrect_comments(question_tag),
            "answers": [
                {
                    "answer_html": answer_html,
                    "answer_weight": 100 if correct else 0
                } for correct, answer_html in answers
            ]
        }
        return question, resources


class MultipleChoiceProcessor(MultipleCommonProcessor):
    question_type = 'multiple_choice_question'
    num_correct = 1

    def process(self, question_tag, markdown_processor: ResourceExtractor):
        return super().process(question_tag, markdown_processor)


class MultipleAnswersProcessor(MultipleCommonProcessor):
    question_type = 'multiple_answers_question'
    num_correct = None

    def process(self, question_tag, markdown_processor: ResourceExtractor):
        return super().process(question_tag, markdown_processor)


class MatchingProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        pairs = question_tag.select('pair')
        matches = []
        for pair in pairs:
            answer_left, answer_right = pair.select('left')[0], pair.select('right')[0]
            matches.append((answer_left.string.strip(), answer_right.string.strip()))

        distractors = question_tag.select('distractors')
        distractor_text = get_text_contents(distractors[0]).strip() if len(distractors) > 0 else None

        question_text, resources = markdown_processor(get_text_contents(
            question_tag,
            ["pair", "distractors", "correct-comments", "incorrect-comments"]
        ))

        question = {
            "question_text": question_text,
            "question_type": 'matching_question',
            "points_possible": get_points(question_tag, default=len(matches)),
            "correct_comments": get_correct_comments(question_tag),
            "incorrect_comments": get_incorrect_comments(question_tag),
            "answers": [
                {
                    "answer_match_left": answer_left,
                    "answer_match_right": answer_right,
                    "answer_weight": 100
                } for answer_left, answer_right in matches
            ],
            "matching_answer_incorrect_matches": distractor_text
        }
        return question, resources


class TextQuestionProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        question_text, resources = markdown_processor(get_text_contents(question_tag))
        question = {
            "question_text": question_text,
            "question_type": 'text_only_question',
        }
        return question, resources


class Parser:
    def __init__(self):
        pass

    def parse(self, string, typ):
        if typ == str:
            return string
        elif typ == int:
            return self.get_int(string)
        elif typ == bool:
            return self.get_bool(string)
        elif typ == list:
            return self.get_list(string)
        elif typ == dict:
            return self.get_dict(string)

    def get_list(self, string):
        items = string.strip().split(',')
        return [cell.strip() for cell in items if cell.strip()]

    def get_bool(self, string):
        # Forgiving boolean parser
        if isinstance(string, bool):
            return string

        if string.lower() == "true":
            return True
        elif string.lower() == "false":
            return False
        else:
            raise ValueError(f"Invalid boolean value: {string}")

    def get_dict(self, string):
        # Assumes the string is a comma-separated list of key-value pairs
        # Example: "key1=value1, key2=value2 "
        return dict(cell.strip().split('=') for cell in string.split(',') if cell.strip())

    def get_int(self, string):
        return int(string)


class OverrideParser:
    def __init__(self, date_formatter, parser):
        self.date_formatter = date_formatter
        self.parser = parser

    def parse(self, override_tag: Tag):
        override = {
            "type": "override",
            "settings": {},
            "sections": [],
            "students": [],
            "assignments": []
        }
        for tag in override_tag.find_all():
            if tag.name == "section":
                override["sections"].append(get_text_contents(tag))
            elif tag.name == "student":
                override["students"].append(get_text_contents(tag))
            elif tag.name == "assignment":
                override["assignments"].append(self.parse_assignment_tag(tag))
        return override

    def parse_assignment_tag(self, tag):
        settings = {
            "title": tag["title"],
        }
        adder = AttributeAdder(settings, tag, self.parser)
        adder("available_from", new_name="unlock_at", formatter=self.date_formatter)
        adder("due_at", formatter=self.date_formatter)
        adder("available_to", new_name="lock_at", formatter=self.date_formatter)
        return settings


class ModuleParser:
    def __init__(self, parser):
        self.parser = parser

    def parse_module_settings(self, module_tag):
        settings = {
            "name": module_tag["title"],
            "position": module_tag["position"],
        }
        AttributeAdder(settings, module_tag, self.parser)("published", False, bool)
        return settings

    def parse(self, module_tag: Tag):
        module = {
            "type": "module",
            "name": module_tag["title"],
            "settings": self.parse_module_settings(module_tag),
            "items": []
        }
        for item_tag in module_tag.find_all():
            module["items"].append(self.parse_module_item(item_tag))
        return module

    casing = {
        "file": "File",
        "page": "Page",
        "discussion": "Discussion",
        "assignment": "Assignment",
        "quiz": "Quiz",
        "subheader": "SubHeader",
        "externalurl": "ExternalUrl",
        "externaltool": "ExternalTool"
    }

    def parse_module_item(self, tag: Tag):
        item = {
            "title": tag["title"],
            "type": self.casing[tag.name],
        }

        adder = AttributeAdder(item, tag, self.parser)
        adder("position", int)
        adder("indent", int)
        adder("page_url")
        adder("external_url")
        adder("new_tab", True, bool)
        adder("completion_requirement")
        adder("iframe")
        adder("published", False, bool)
        return item


class QuizParser:
    question_processors = {
        "multiple-choice": MultipleChoiceProcessor(),
        "multiple-answers": MultipleAnswersProcessor(),
        "true-false": TrueFalseProcessor(),
        "multiple-tf": MultipleTrueFalseProcessor(),
        "matching": MatchingProcessor(),
        "text": TextQuestionProcessor()
    }

    def __init__(self, markdown_processor: ResourceExtractor, group_indexer, date_formatter, parser):
        self.markdown_processor = markdown_processor
        self.group_indexer = group_indexer
        self.date_formatter = date_formatter
        self.parser = parser

    def parse(self, quiz_tag: Tag):
        quiz = {
            "type": "quiz",
            "name": quiz_tag["title"],
            "questions": [],
            "resources": []
        }
        quiz.update(self.parse_quiz_settings(quiz_tag))

        for tag in quiz_tag.find_all():
            if tag.name == "question":
                question, res = self.parse_question(tag)
                quiz["resources"].extend(res)
                # if question is a  list of questions, add them all
                if isinstance(question, list):
                    quiz["questions"].extend(question)
                else:
                    quiz["questions"].append(question)
            elif tag.name == "description":
                description, res = self.markdown_processor(get_text_contents(tag))
                quiz["resources"].extend(res)
                quiz["description"] = description
        return quiz

    def parse_quiz_settings(self, settings_tag):
        settings = {"title": settings_tag["title"]}

        adder = AttributeAdder(settings, settings_tag, self.parser)

        adder("quiz_type", "assignment")
        adder("assignment_group", None, "assignment_group_id", formatter=self.group_indexer)
        adder("time_limit", None, typ=int)
        adder("shuffle_answers", False, typ=bool)
        adder("hide_results", typ=str)
        adder("show_correct_answers", True, typ=bool)
        adder("show_correct_answers_last_attempt", False, typ=bool)
        adder("show_correct_answers_at", None, formatter=self.date_formatter)
        adder("hide_correct_answers_at", None, formatter=self.date_formatter)
        adder("allowed_attempts", -1, typ=int)
        adder("scoring_policy", "keep_highest")
        adder("one_question_at_a_time", False, typ=bool)
        adder("cant_go_back", False, typ=bool)
        adder("available_from", None, "unlock_at", formatter=self.date_formatter)
        adder("due_at", None, formatter=self.date_formatter)
        adder("available_to", None, "lock_at", formatter=self.date_formatter)
        adder("access_code")
        adder("published", False, typ=bool)
        adder("one_time_results", False, typ=bool)

        return settings

    def parse_question(self, question_tag: Tag):
        processor = self.question_processors[question_tag["type"]]
        return processor.process(question_tag, self.markdown_processor)


class AssignmentParser:
    def __init__(self, markdown_processor: ResourceExtractor, group_indexer, date_formatter, parser):
        self.markdown_processor = markdown_processor
        self.group_indexer = group_indexer
        self.date_formatter = date_formatter
        self.parser = parser

    def parse(self, assignment_tag):
        assignment = {
            "type": "assignment",
            "name": assignment_tag["title"],
            "resources": [],
            "settings": self.parse_assignment_settings(assignment_tag)
        }

        for tag in assignment_tag.find_all():
            if tag.name == "description":
                contents = get_text_contents(tag)
                description, res = self.markdown_processor(contents)
                assignment["settings"]["description"] = description
                assignment["resources"].extend(res)

        assignment["name"] = assignment["name"]
        return assignment

    def parse_assignment_settings(self, settings_tag):
        settings = {"name": settings_tag["title"]}

        adder = AttributeAdder(settings, settings_tag, self.parser)
        adder("allowed_attempts", formatter=lambda x: -1 if x == "not_graded" else x),
        adder("allowed_extensions", "", typ=list),
        adder("annotatable_attachment_id"),
        adder("assignment_group", new_name="assignment_group_id", formatter=self.group_indexer),
        adder("assignment_overrides"),
        adder("automatic_peer_reviews", False, typ=bool),
        adder("available_from", new_name="unlock_at", formatter=self.date_formatter),
        adder("available_to", new_name="lock_at", formatter=self.date_formatter),
        adder("due_at", formatter=self.date_formatter),
        adder("external_tool_tag_attributes", "", typ=dict),
        adder("final_grader_id"),
        adder("grade_group_students_individually", False, typ=bool),
        adder("grading_standard_id"),
        adder("grading_type", "points"),
        adder("grader_comments_visible_to_graders", False, typ=bool),
        adder("grader_count"),
        adder("grader_names_visible_to_final_grader", False, typ=bool),
        adder("graders_anonymous_to_graders", False, typ=bool),
        adder("group_category", new_name="group_category_id"),
        adder("hide_in_gradebook", False, typ=bool),
        adder("integration_data"),
        adder("moderated_grading", False, typ=bool),
        adder("notify_of_update", False, typ=bool),
        adder("omit_from_final_grade", False, typ=bool),
        adder("only_visible_to_overrides", False, typ=bool),
        adder("peer_reviews", False, typ=bool),
        adder("points_possible"),
        adder("position"),
        adder("published", False, typ=bool),
        adder("quiz_lti"),
        adder("submission_types", "none", typ=list),
        adder("turnitin_enabled", False, typ=bool),
        adder("turnitin_settings"),
        adder("vericite_enabled", False, typ=bool)

        return settings


class PageParser:
    def __init__(self, markdown_processor: ResourceExtractor, date_formatter, parser):
        self.markdown_processor = markdown_processor
        self.date_formatter = date_formatter
        self.parser = parser

    def parse_page_settings(self, page_tag):
        settings = {
            "type": "page",
            "title": page_tag["title"],
            "body": "",
        }
        adder = AttributeAdder(settings, page_tag, self.parser)
        adder("editing_roles", "teachers")
        adder("notify_of_update", False, typ=bool)
        adder("published", False, typ=bool)
        adder("front_page", False, typ=bool)
        adder("publish_at", formatter=self.date_formatter)
        return settings

    def parse(self, page_tag):
        page = {
            "type": "page",
            "name": page_tag["title"],
            "settings": self.parse_page_settings(page_tag),
            "resources": []
        }
        contents = get_text_contents(page_tag)
        body, res = self.markdown_processor(contents)
        page["settings"]["body"] = body
        page["resources"].extend(res)
        return page


class DocumentParser:
    def __init__(self, path_to_resources: Path, path_to_canvas_files: Path, markdown_processor: ResourceExtractor,
                 time_zone: str,
                 group_identifier=lambda x: 0):
        self.path_to_resources = path_to_resources
        self.path_to_files = path_to_canvas_files
        self.markdown_processor = markdown_processor
        self.date_formatter = lambda x: make_iso(x, time_zone)

        self.jinja_env = Environment()
        # This enables us to use the zip function in template documents

        self.jinja_env.globals.update(zip=zip, split_list=lambda sl: [s.strip() for s in sl.split(';')])

        parser = Parser()

        self.element_processors = {
            "quiz": QuizParser(self.markdown_processor, group_identifier, self.date_formatter,parser),
            "assignment": AssignmentParser(self.markdown_processor, group_identifier, self.date_formatter, parser),
            "page": PageParser(self.markdown_processor, self.date_formatter, parser),
            "module": ModuleParser(parser),
            "override": OverrideParser(self.date_formatter, parser)
        }

    def parse(self, text):
        soup = BeautifulSoup(text, "html.parser")
        document = []
        tag: Tag
        for tag in soup.children:
            parser = self.element_processors.get(tag.name, None)
            if parser:
                templates = parser.parse(tag)
                if not isinstance(templates, list):
                    templates = [templates]
                templates = [order_elements(template) for template in templates]
        return document




from datetime import datetime

from bs4 import Tag

from .attributes import parse_settings, Attribute, parse_date, parse_bool
from ..resources import ResourceManager, CanvasResource
from ..util import retrieve_contents


class AnnouncementTagProcessor:
    def __init__(self, resources: ResourceManager):
        self._resources = resources

    def __call__(self, announcement_tag: Tag):
        # Note - it appears you can edit announcement that haven't published yet
        # Not sure if you can edit announcements that have published already
        # Gordon Bean (Jan 17, 2025)

        fields = [
            Attribute('title', required=True),
            Attribute('is_announcement', True, parser=parse_bool),
            Attribute('publish_date', required=True, new_name='delayed_post_at', parser=parse_date,
                      default=datetime.now().isoformat()),
        ]

        # https://canvas.instructure.com/doc/api/discussion_topics.html#method.discussion_topics.create
        # delayed_post_at
        # is_section_specific
        # permissions.reply
        # published
        # specific_sections

        settings = {
            "type": "discussion_topics",
            "message": retrieve_contents(announcement_tag)
        }

        settings.update(parse_settings(announcement_tag, fields))

        announcement = CanvasResource(
            type='announcement',
            name=settings['title'],
            data=settings,
        )
        self._resources.add_resource(announcement)



from bs4 import Tag

from .attributes import parse_settings, Attribute, parse_date, parse_int
from ..resources import ResourceManager, CanvasResource, get_key


class OverrideTagProcessor:
    def __init__(self, resources: ResourceManager):
        self._resources = resources

    def __call__(self, override_tag: Tag):
        fields = [
            Attribute('available_from', parser=parse_date, new_name='unlock_at'),
            Attribute('available_to', parser=parse_date, new_name='lock_at'),
            Attribute('due_at', parser=parse_date),
            Attribute('late_due', parser=parse_date),
            Attribute('title'),  # required if using student IDs
            Attribute('section_id', new_name='course_section_id', required=True, parser=parse_int),
            # TODO - support student IDs also
            Attribute('title', new_name='assignment_name', required=True),  # name of assignment to modify with this override
            Attribute('type', new_name='rtype', required=True)
        ]

        settings = {
            "type": "override",
        }

        settings.update(parse_settings(override_tag, fields))

        settings['name'] = f"{settings['assignment_name']}|{settings['course_section_id']}"
        settings['assignment_id'] = get_key(settings['rtype'], settings['assignment_name'], 'id')

        assignment = CanvasResource(
            type='override',
            name=settings['name'],
            data=settings
        )
        self._resources.add_resource(assignment)
